/*************************************************************************

    File: dpurc-sem.pl
    Copyright (C) 2008 Johannes Dellert

    This file is part of LARA, version 1.0 (August 2008).

*************************************************************************/

:- use_module(comsemPredicates,[infix/0,
                                prefix/0,
                                printRepresentations/1,
                                memberList/2,
                                reverseList/2,
                                selectFromList/3,
				                compose/3]).

:- use_module(betaConversion,[betaConvert/2]).

:- use_module(alphaConversion,[alphabeticVariants/2]).

:- [dpurc].

:- [dpurc-gr-eo].

:- [eo-dict-sem].

/*========================================================================
   Retrieve all the LFs for an input sentence
========================================================================*/

syn_sem_parse(Input, Sems) :-
  setof(Sem,syn_sem_single_parse(Input,Sem),AllSems),
  filterAlphabeticVariants(AllSems,Sems).


/*========================================================================
   Retrieve a single LF for an input sentence
========================================================================*/

syn_sem_single_parse(Input, Sem) :-
  parse(Input,[Head]),
  sem_single_parse(Head,Sem).


/*========================================================================
   Retrieve all the LFs for a dependency structure
========================================================================*/

sem_parse(Head,Sems) :-
  setof(Sem,sem_single_parse(Head,Sem),AllSems),
  filterAlphabeticVariants(AllSems,Sems).


/*========================================================================
   Retrieve a single LF for a dependency structure
========================================================================*/

sem_single_parse(Head, Converted) :-
  sem_proc(Head, Sem),
  betaConvert(Sem,Converted),
  % suppress non-sentential LFs
  \+ lambda_or_app_occurrence(Converted) .


/*========================================================================
   Compute LFs for a dependency structure
========================================================================*/

sem_proc([_,_,_,_,_,Dependents,_,Cat,Lemma|_], Sem) :-
  Dependents = [],
  semLex(Cat, Lemma, Sem).

sem_proc([_,_,_,_,_,Dependents,_,Cat,Lemma|_],Sem) :-
  compute_dep_sem(Dependents, [], DepSems),
  permutation(DepSems, InverseDepSemsVariant),
  %must be checked in inverse order
  \+ violates_precedence(Cat,InverseDepSemsVariant),
  reverseList(InverseDepSemsVariant, DepSemsVariant),
  semLex(Cat,Lemma,HeadSem),
  apply_directionally(Cat, HeadSem, DepSemsVariant, Sem).


/*========================================================================
   Compute semantic values for dependents
========================================================================*/

compute_dep_sem([],DepSems,DepSems).

compute_dep_sem([Dep|Dependents], DepSems, ReturnDepSems) :-
  % check necessary because Dependents is an open list; also: base case
  \+ var(Dep),
  sem_proc(Dep, DepSem),
  Dep = [Label|_],
  append(DepSems,[Label:DepSem],NewDepSems),
  compute_dep_sem(Dependents, NewDepSems, ReturnDepSems).


/*============================================================================
   Find precedence constraint violations in (inverse) orders of dependent LFs 
==============================================================================*/

violates_precedence(HeadCat,[First:_|DepSemsSequence]) :- 
  ( precedence(HeadCat,First,Preceded),
    memberList(DepSemsSequence,Preceded:_));
  violates_precedence(HeadCat,DepSemsSequence).


/*========================================================================
   Combine semantics of a head and its dependents
   according to the directionality constraints
========================================================================*/

apply_directionally(_,HeadSem,[],HeadSem).

apply_directionally(_,unknown,_,unknown).

apply_directionally(HeadCat,HeadSem,[_:unknown|OtherDepSems],FinalSem) :-
  apply_directionally(HeadCat,HeadSem,OtherDepSems,FinalSem). 

apply_directionally(HeadCat,HeadSem,[Label:DepSem|OtherDepSems],FinalSem) :-
  directional(HeadCat:HeadSem,Label,_:DepSem,NextSem),
  apply_directionally(HeadCat,NextSem,OtherDepSems,FinalSem).   


/*========================================================================
   Helper predicates to generate all permutations
========================================================================*/

select(X,[X|Xs],Xs).
select(X,[Y|Ys],[Y|Zs]) :- select(X,Ys,Zs).

permutation(Xs,[Z|Zs]) :- select(Z,Xs,Ys), permutation(Ys,Zs).
permutation([],[]).


/*========================================================================
   Find non-sentential LFs in B&B notation
========================================================================*/

lambda_or_app_occurrence(Term) :-
  \+ var(Term),
  functor(Term,Name,Arity),
  (
    Name = lam;
    Name = app;
    (
      arg(_,Term,_),
      lambda_or_app_occurrence_args(Term, 1, Arity)
    )
  ).

lambda_or_app_occurrence_args(Term, ArgID, Arity) :-
  arg(ArgID,Term,Arg),
  (
    lambda_or_app_occurrence(Arg);
    (
      NextArgID is ArgID + 1,
      NextArgID =< Arity,
      lambda_or_app_occurrence_args(Term,NextArgID, Arity)
    )
  ).

/*========================================================================
   Filter Alphabetic Variants (Copyright Blackburn & Bos)
========================================================================*/

filterAlphabeticVariants(L1,L2):-
   selectFromList(X,L1,L3),
   memberList(Y,L3),
   alphabeticVariants(X,Y), !,
   filterAlphabeticVariants(L3,L2).

filterAlphabeticVariants(L,L).


/*========================================================================
   Collect and print all the LFs for a sentence
========================================================================*/

try_all_sem(List) :- 
  syn_sem_parse(List, Sems),
  printRepresentations(Sems).
  

/*========================================================================
   Collect and print all the dependency structures for a sentence
   together with their interpretations (if any)
========================================================================*/

try_sem(List) :- 
  write_list(List),nl,
  parse(List,[Head]),
  write_dep(Head),
  sem_parse(Head, Formulas),
  printRepresentations(Formulas),
	nl,
  nl,
	fail.

try_sem(_) :- write('No (more) parses.'), nl.
