% dpurc.pl - Johannes Dellert 2008    johannes.dellert@gmx.de
%
% An extended non-projective dependency parser in plain Prolog,
% allowing constraints on headedness and projectivity of relations.

% The main architecture and concept of this parser as well as the output routines
% build on a very simple example parser by M. Covington 2003: http://www.ai.uga.edu/mc

% Michael Covington has made important suggestions on how to improve his parser 
% in an additional paper, most of which have been implemented here.

%
% DPURC - Dependency Parser with User-definable Relation Constraints
%

% We maintain 2 lists, WordList and HeadList.
% WordList contains all words prior to the current one.
% HeadList contains all words that do not have heads.
% Eventually HeadList will have only one word in it.

% In WordList and HeadList, each word is represented as a "node"
% of the form:
% [Label,Number,Start,End,Projective,Dependents,Realized|Features].
% Features are as defined by lexical entries in the grammar;
% Meaning of the other components:
% Label, the relational label for the node's dependency
% Number, a unique identifier, also starting index of the word
% Start, at which position the constituent dependent on this word starts
% End, at which position the constituent dependent on this word ends
% Projective, whether the constituent dependent on this word is cohesive
% Dependents, an open list (initially empty) of dependents.
% Realized, an open list (initially empty) of realized dependents

% We build the list of Dependents for each word by instantiating
% its initially uninstantiated tail.  That way, we can work on the
% word in either list (WordList or HeadList) and its representation
% in the other list will change with it, because it contains the
% same variable.


% parse(+InputList,-Result)
%  Parses a list of words, giving a HeadList.

parse(InputList,Result) :-
   parse_loop(1,InputList,[],[],Result).


% parse_loop(+Count,+InputList,+WordList,+HeadList,-Result)
%  Called by parse/2, to loop through the list.

parse_loop(Count,[Word|InputList],WordList,HeadList,Result) :-
   word(Word,WordFeatures),                        % Look up the next word
   Start is Count,
   End is Count + 1,
   Node = [_,Count, Start, End, yes, _, _ | WordFeatures],               % Build a node for it
   parse_node(Node,WordList,NewWordList,HeadList,NewHeadList,NewNode),   % Try to attach it and modify it
   NewCount is Count + 1,
   ExtendedWordList = [NewNode|NewWordList],                    % Add modified Node to WordList
   parse_loop(NewCount,InputList,ExtendedWordList,NewHeadList,Result).

parse_loop(_,[],_,Result,Result).
   % No more words to parse; so Result := HeadList.



% parse_node(+Node,+WordList,+HeadList,-NewHeadList,-NewNode)
%  Try to attach Node to the dependency tree.
%  HeadList gets modified here; WordList does not
%  (except by instantiation).
%  NewNode contains node as modified by insertion

parse_node(Node,[],[],[],[Node],Node) :- !.
   % If this is the first word, just add it to HeadList and return it identically.

parse_node(Node,WordList,NewWordList,HeadList,NewHeadList,NewNode) :-
   try_inserting_as_head(Node,HeadList,HeadList2,NewNode),
   try_inserting_as_dependent(NewNode,WordList,HeadList2,NewHeadList,NewWordList,no).



% try_inserting_as_head(+Node,+HeadList,-NewHeadList,-NewNode)
%  Try to insert Node as the head above a word presently in HeadList.
%  This entails deleting an element from HeadList.
%  NewNode contains a modified version of Node if successful

try_inserting_as_head(Node,[],[],Node).
   % No more elements of HeadList to look at.

try_inserting_as_head(Node,[Head|HeadList],NewHeadList,NextCore) :-
   % Insert Node above Head, and remove Head from HeadList.
   Node = [NodeLabel,NodePos,_,_,_,D,R|NodeFeatures],
   Head = [_,HeadPos,HeadStart,HeadEnd,HeadProjectivity,HeadD,HeadR|HeadFeatures],
   dh(HeadFeatures,NodeFeatures,Headedness,Projectivity,Uniqueness,Label),  % Head is the dependent here
   check_headedness(HeadPos, NodePos, Headedness),
   check_projectivity(HeadProjectivity, Projectivity),
   calculateProjectivityData(Node,Head,NewCoreProjectivity, NewCoreStart, NewCoreEnd),
   % HeadProjectivity must be changed to "fixed" here
   determineDependentHeadProjectivity(Projectivity,ThreeValuedProjectivity),
   % this must be the node handed on to the next step if successful
   NewCore = [NodeLabel,NodePos,NewCoreStart,NewCoreEnd,NewCoreProjectivity,D,R|NodeFeatures],
   % dependent has to be defined with ThreeValuedProjectivity
   % THE THREE-VALUED PROJECTIVITY DOES NOT SHOW UP IN THE WORD LIST RELEVANT LATER!
   NewHead = [Label,HeadPos,HeadStart,HeadEnd,ThreeValuedProjectivity,HeadD,HeadR|HeadFeatures],
   add_item(D,NewHead),
   try_adding_realized_item(R, Label, Uniqueness),
   % Now recurse down HeadList and see if it can be done again.
   % (Multiple elements of HeadList may be dependents of Node.)
   try_inserting_as_head(NewCore,HeadList,NewHeadList,NextCore).
   
try_inserting_as_head(Node,[Head|HeadList],[Head|NewHeadList],NewNode) :-
   % Couldn't use Head, so skip it.
   try_inserting_as_head(Node,HeadList,NewHeadList,NewNode).

try_adding_realized_item(_,_, no).

try_adding_realized_item(R, Label, yes) :- 
   \+ contains_item(R,Label),
   add_item(R,Label).


% determineDependentHeadProjectivity(+Projectivity,-ThreeValuedProjectivity)
%  Determines the projectivity constraint on a dependent that has previously been a head

determineDependentHeadProjectivity(no,no).

determineDependentHeadProjectivity(yes,fixed).


% try_inserting_as_dependent(+Node,+WordList,+HeadList,-NewHeadList,-NewWordList,+Attached)
%  Try to insert Node as a dependent of a word already encountered,
%  i.e., of an element of WordList.
%  Such an operation would change the word leading to NewWordList.
%  Alternatively, add Node to HeadList, leading to NewHeadList.
%  Attached keeps track of the Node has been attached during recursion

% structures already dependent on something may be extended non-projectively
% SOLUTION: three states of Projectivity: "yes", "no" and "fixed" (mustn't be changed to "no" anymore)

try_inserting_as_dependent(Node,[Word|WordList],HeadList,NewHeadList,[NewWord|NewWordList],_) :-
   Word = [WordLabel,WordPos,_,_,WordProjectivity,D, R|WordFeatures],
   Node = [Label,NodePos,_,_,NodeProjectivity,_,_|NodeFeatures], %Node is the dependent here
   dh(NodeFeatures,WordFeatures,Headedness,ProjectivityConstraint,Uniqueness,Label),
   check_headedness(NodePos, WordPos, Headedness),
   check_projectivity(NodeProjectivity,ProjectivityConstraint),
   calculateProjectivityData(Word,Node,NewWordProjectivity, NewWordStart, NewWordEnd),
   NewWord = [WordLabel,WordPos,NewWordStart,NewWordEnd,NewWordProjectivity,D,R|WordFeatures],
   check_projectivity_compatibility(WordProjectivity,NewWordProjectivity),
   add_item(D,Node),
   try_adding_realized_item(R, Label, Uniqueness),
   try_inserting_as_dependent(Node,WordList,HeadList,NewHeadList,NewWordList,yes).

try_inserting_as_dependent(Node,[Word|WordList],HeadList,NewHeadList,[Word|NewWordList],Attached) :-
   % Couldn't use Word, so skip it.
   try_inserting_as_dependent(Node,WordList,HeadList,NewHeadList,NewWordList,Attached).

   % couldn't attach it, so add it to HeadList
try_inserting_as_dependent(Node,[],HeadList,[Node|HeadList],[],no).

   % do not add Node to HeadList if Node could be attached before
try_inserting_as_dependent(_,[],HeadList,HeadList,[],yes).



% check_headedness(+DepPosition,+HeadPosition,+Headedness)
%  take headedness information of dependency rules into account when deciding whether a relation can be established
%  free_order means that the positions of head and dependent do not matter
%  head_final means the position of the dependent must be earlier than the position of the head
%  head_initial means the position of the head must be earlier than the position of the dependent

check_headedness(_,_,Headedness) :-
    Headedness = free_order.

check_headedness(DepPosition,HeadPosition,Headedness) :-
    DepPosition < HeadPosition,
    Headedness = head_final.

check_headedness(DepPosition,HeadPosition,Headedness) :-
    DepPosition > HeadPosition,
    Headedness = head_initial.

% check_projectivity_compatibility(+Pro1,+Pro2)
%  determine whether a change in projectivity violates the constraints

check_projectivity_compatibility(fixed,yes).

check_projectivity_compatibility(yes,yes).

check_projectivity_compatibility(no,no).


% check_projectivity(+Projectivity,+Constraint)
%  take projectivity information of dependency rules into account when deciding whether a relation can be established 
%  two possible values for Projectivity: "yes" and "no"
%  will accept if projectivity matches constraint

check_projectivity(fixed,yes).

check_projectivity(yes,yes).

check_projectivity(_,no).


% calculateProjectivityData(+Core, +Dependent, -NewCoreProjectivity, -NewCoreStart, -NewCoreEnd)
%  determine the projectivity data of the core after the dependent has been appended
%  will determine start and end positions of the extended constituent
%  and whether it can still be considered projective

calculateProjectivityData(Core, Dependent, NewCoreProjectivity, NewCoreStart, NewCoreEnd) :-
   Core = [_,_,CoreStart,CoreEnd,CoreProjectivity|_],
   Dependent = [_,_,DepStart,DepEnd,DepProjectivity|_],
   calculateProjectivity(CoreStart,CoreEnd,DepStart,DepEnd,CoreProjectivity,DepProjectivity,NewCoreProjectivity),
   calculateNewBoundaries(CoreStart,CoreEnd,DepStart,DepEnd,NewCoreStart,NewCoreEnd).


% calculateProjectivity(+CoreStart, +CoreEnd, +DepStart, +DepEnd, +CoreProj,+DepProj,-Result)
%  determine the projectivity of a newly created constituent
%  tries to connect the positions of the two constituents linked to a new coherent area
%  result is "no" if this is impossible


calculateProjectivity(_,CoreEnd,DependentStart,_,CoreProjectivity,DependentProjectivity,yes) :-
    DependentProjectivity = yes,
    CoreProjectivity = yes,
    CoreEnd = DependentStart, !.

calculateProjectivity(CoreStart,_,_,DependentEnd,CoreProjectivity,DependentProjectivity,yes) :-
    DependentProjectivity = yes,
    CoreProjectivity = yes,
    CoreStart = DependentEnd, !.

calculateProjectivity(_,CoreEnd,DependentStart,_,CoreProjectivity,DependentProjectivity,yes) :-
    DependentProjectivity = fixed,
    CoreProjectivity = yes,
    CoreEnd = DependentStart, !.

calculateProjectivity(CoreStart,_,_,DependentEnd,CoreProjectivity,DependentProjectivity,yes) :-
    DependentProjectivity = fixed,
    CoreProjectivity = yes,
    CoreStart = DependentEnd, !.

calculateProjectivity(_,_,_,_,_,_,no).


% calculateProjectivityData(+CoreStart, +CoreEnd, +DepStart, +DepEnd, -Start,-End)
%  determine the range of a constituent formed by Core and Dep
%  returns start and end of the common range, whether coherent or not

calculateNewBoundaries(CoreStart,CoreEnd,DependentStart,DependentEnd,CoreStart, DependentEnd) :-
    CoreEnd < DependentStart, !.

calculateNewBoundaries(CoreStart,CoreEnd,DependentStart,DependentEnd,DependentStart, CoreEnd) :-
    CoreStart > DependentEnd, !.

calculateNewBoundaries(CoreStart,CoreEnd,DependentStart,DependentEnd,CoreStart, DependentEnd) :-
    CoreEnd = DependentStart, !.

calculateNewBoundaries(CoreStart,CoreEnd,DependentStart,DependentEnd,DependentStart, CoreEnd) :-
    CoreStart = DependentEnd.

    

%
% LIST UTILITIES
%


% add_item(?OpenList,+Item)
%   adds Item just before the tail of OpenList

add_item(X,Item) :- var(X), !, X = [Item|_].

add_item([_|X],Item) :- add_item(X,Item).

% contains_item(?OpenList,+Item)
%   checks whether Item is already present on OpenList

contains_item(OpenList, Item) :- \+ var(OpenList), OpenList = [Item|_].

contains_item(OpenList, Item) :- OpenList = [_|Rest], \+ var(Rest), contains_item(Rest,Item).



% Output utilities


write_list([First|Rest]) :-
   var(First),
   !,
   write('_ '),
   write_list(Rest).

write_list([First|Rest]) :-
   write(First),
   write(' '),
   write_list(Rest).

write_list([]) :-
   nl.



write_dep(Node) :-
   write_dep(Node,0).

write_dep([Label,N,_,_,_,Dependents,_|Rest],Indentation) :-
   tab(Indentation),
   write(Label), write(' '), write(N), write(' '), write_list(Rest),
   NewIndentation is Indentation + 3,
   write_dep_open_list(Dependents,NewIndentation).

write_dep_open_list(X,_) :-
   var(X),
   !.

write_dep_open_list([First|Rest],N) :-
   write_dep(First,N),
   write_dep_open_list(Rest,N).

   



% tests

try(List) :- write_list(List),nl,
             parse(List,[Head]),
             write_dep(Head),
	     nl,
	     fail.

try(_) :- write('No (more) parses.'), nl.







