/*************************************************************************

    File: helpemaLara.pl
    Copyright (C) 2008 Johannes Dellert

    This file is part of LARA, version 1.0 (August 2008).

    Original file: helpfulCurt.pl
    Copyright (C) 2004,2005,2006 Patrick Blackburn & Johan Bos

    That file was part of BB1, version 1.3 (November 2006).


*************************************************************************/


:- module(lara,[lara/0,infix/0,prefix/0]).

:- use_module(callInference,[callTP/3,
                             callTPandMB/6]).

:- use_module(readLine,[readLine/1]).

:- use_module(comsemPredicates,[infix/0,
                                prefix/0,
                                memberList/2,
                                compose/3,
				                selectFromList/3,
				                printRepresentations/1]).

:- use_module(modelChecker2,[satisfy/4]).

:- use_module(backgroundKnowledge,[backgroundKnowledge/2]).

:- use_module(elimEquivReadings,[elimEquivReadings/2]).

:- use_module(laraPredicates,[laraHelp/0,
                              laraOutput/1,
                              updateReadings/1,
                              updateModels/1,
                              updateHistory/1,
                              clearHistory/0,
                              list2string/2,
                              selectReadings/3]).

:- [dpurc-sem].


/*========================================================================
   Dynamic Predicates
========================================================================*/

:- dynamic history/1, readings/1, models/1.

history([]).
readings([]).
models([]).


/*========================================================================
   Start Lara
========================================================================*/

lara:- 
   laraTalk(run).


/*========================================================================
   Control
========================================================================*/

laraTalk(quit).

laraTalk(run):-
   readLine(Input),
   laraUpdate(Input,LarasMoves,State), 
   laraOutput(LarasMoves),
   laraTalk(State).


/*========================================================================
   Update Lara's Information State
========================================================================*/

laraUpdate([],[clarify],run):- !.

laraUpdate([gxis],[gxis],quit):- !,
   updateReadings([]),
   updateModels([]),
   clearHistory.

laraUpdate([denove],[],run):- !,
   updateReadings([]),
   updateModels([]),
   clearHistory.

laraUpdate([helpo],[],run):- !,
   laraHelp.

laraUpdate([infikso],[],run):- !,
   infix.

laraUpdate([prefikso],[],run):- !,
   prefix.

laraUpdate([elektu,X],[],run):-
   number(X),
   readings(R1),
   selectReadings(X,R1,R2), !,
   updateReadings(R2),
   models(M1),
   selectReadings(X,M1,M2),
   updateModels(M2).

laraUpdate([resumo],[],run):-
   readings(Readings),
   elimEquivReadings(Readings,Unique),
   updateReadings(Unique),
   updateModels([]).

laraUpdate([scio],[],run):-
   readings(R), 
   findall(K,(memberList(F,R),backgroundKnowledge(F,K)),L),
   printRepresentations(L).
   
laraUpdate([interpretoj],[],run):- !,
   readings(R),
   printRepresentations(R).

laraUpdate([modeloj],[],run):- !,
   models(M),
   printRepresentations(M).

laraUpdate([historio],[],run):- !,
   history(H),
   printRepresentations(H).

laraUpdate(Input,Moves,run):-
   syn_sem_parse(Input,Readings), !,
   updateHistory(Input),
   (
      Readings=[que(X,R,S)|_],
      models(OldModels),
      answerQuestion(que(X,R,S),OldModels,Moves)
   ;  
      \+ Readings=[que(_,_,_)|_],
      consistentReadings(Readings,[]-ConsReadings,[]-Models),
      (
         ConsReadings=[],
         Moves=[contradiction]
      ;
         \+ ConsReadings=[],
         informativeReadings(ConsReadings,[]-InfReadings),   
         (
            InfReadings=[],
            Moves=[obvious]
         ;  
            \+ InfReadings=[],
            Moves=[accept]
         ),
         combine(ConsReadings,CombinedReadings), 
         updateReadings(CombinedReadings),
         updateModels(Models)
      )
   ).

laraUpdate(_,[noparse],run).


/*========================================================================
   Combine New Utterances with History
========================================================================*/

combine(New,New):-
   readings([]).

combine(Readings,Updated):-
   readings([Old|_]),
   findall(and(Old,New),memberList(New,Readings),Updated).


/*========================================================================
   Select Consistent Readings
========================================================================*/

consistentReadings([],C-C,M-M).

consistentReadings([New|Readings],C1-C2,M1-M2):-
   readings(Old),
   (
      consistent(Old,New,Model), !,
      consistentReadings(Readings,[New|C1]-C2,[Model|M1]-M2) 
   ;
      consistentReadings(Readings,C1-C2,M1-M2) 
   ).


/*========================================================================
   Consistency Checking calling Theorem Prover and Model Builder
========================================================================*/

consistent([Old|_],New,Model):-
   DomainSize=15,
   callTPandMB(not(and(Old,New)),and(Old,New),DomainSize,Proof,Model,_),
   \+ Proof=proof, Model=model([_|_],_).

consistent([],New,Model):-
   DomainSize=15,
   callTPandMB(not(New),New,DomainSize,Proof,Model,_),
   \+ Proof=proof, Model=model([_|_],_).


/*========================================================================
   Select Informative Readings
========================================================================*/

informativeReadings([],I-I).

informativeReadings([New|L],I1-I2):-
   readings(Old),
   (
      informative(Old,New), !,
      informativeReadings(L,[New|I1]-I2) 
   ;
      informativeReadings(L,I1-I2) 
   ).


/*========================================================================
   Informativity Checking calling Theorem Prover
========================================================================*/

informative([Old|_],New):-
   DomainSize=15,
    callTPandMB(not(and(Old,not(New))),and(Old,not(New)),DomainSize,Proof,Model,_),
   \+ Proof=proof, Model=model([_|_],_).  

informative([],New):-
   DomainSize=15,
    callTPandMB(not(not(New)),not(New),DomainSize,Proof,Model,_),
   \+ Proof=proof, Model=model([_|_],_).


/*========================================================================
    Answer Questions
========================================================================*/

answerQuestion(que(X,R,S),Models,Moves):-
   (
      Models=[Model|_],
      satisfy(some(X,S),Model,[],Result), 
      \+ Result=undef,
      !,     
      findall(A,satisfy(S,Model,[g(X,A)],pos),Answers),
      realiseAnswer(Answers,que(X,R,S),Model,String),
      Moves=[sensible_question,answer(String)]
   ;
      Moves=[unknown_answer]
   ).


/*========================================================================
    Realise all answers
========================================================================*/

realiseAnswer([],_,_,'neniu').

realiseAnswer([Value],Q,Model,String):-
   realiseString(Q,Value,Model,String).

realiseAnswer([Value1,Value2|Values],Q,Model,String):-
   realiseString(Q,Value1,Model,String1),
   realiseAnswer([Value2|Values],Q,Model,String2),
   list2string([String1,kaj,String2],String).


/*========================================================================
    Realise a single answer
========================================================================*/

realiseString(que(X,R,S),Value,Model,String):-
   Model = model(_,RelList),
   memberList(Rel,RelList),
   Rel = f(0,Symbol,Value),
   satisfy(eq(Y,Symbol),Model,[g(Y,Value)],pos), !,
   checkAnswer(some(X,and(eq(X,Symbol),and(R,S))),Proof),
   (
      Proof=proof, !,
      list2string([Symbol],String)
   ;
      list2string([eble|Symbol],String)
   ).

realiseString(que(X,R,S),Value,Model,String):-
   Model = model(_,RelList),
   memberList(Rel,RelList),
   Rel = f(1,Symbol,Set),
   memberList(Value,Set),
   compose(Formula,Symbol,[X]),
   satisfy(Formula,Model,[g(X,Value)],pos), !,
   checkAnswer(some(X,and(Formula,and(R,S))),Proof),
   (
      Proof=proof, !,
      list2string([Symbol],String)
   ;
      list2string([eble|Symbol],String)
   ).

realiseString(_,Value,_,Value).


/*========================================================================
   Answer Checking
========================================================================*/

checkAnswer(Answer,Proof):-
   readings([F|_]),
   backgroundKnowledge(F,BK),
   callTP(imp(and(F,BK),Answer),Proof,_).


/*========================================================================
   Info
========================================================================*/

info:-
   format('~n> ---------------------------------------------------------- <',[]),
   format('~n> helpemaLara.pl, de Johannes Dellert                        <',[]),
   format('~n>                                                            <',[]),
   format('~n> ?- lara.                - iniciatu dialogon kun Lara       <',[]),
   format('~n>                                                            <',[]),
   format('~n> Tajpu "helpo" por pli informo pri la trajtoj de Lara       <',[]),
   format('~n> ---------------------------------------------------------- <',[]),
   format('~n~n',[]).



/*========================================================================
   Display info at start
========================================================================*/

:- info.

 
